"""
https://github.github.com/gfm/#block-quotes
"""
import pytest

from .utils import act_and_assert

# pylint: disable=too-many-lines


@pytest.mark.gfm
def test_block_quotes_206x():
    """
    Test case 206:  Here is a simple example:
    """

    # Arrange
    source_markdown = """> # Foo
> bar
> baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[atx(1,3):1:0:]",
        "[text(1,5):Foo: ]",
        "[end-atx::]",
        "[para(2,3):\n]",
        "[text(2,3):bar\nbaz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h1>Foo</h1>
<p>bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_206a():
    """
    Test case 206:  Here is a simple example:
    """

    # Arrange
    source_markdown = """> Foo
> bar
> baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):\n\n]",
        "[text(1,3):Foo\nbar\nbaz::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>Foo
bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_206b():
    """
    Test case 206:  Here is a simple example:
    """

    # Arrange
    source_markdown = """> Foo
bar
baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n]",
        "[para(1,3):\n\n]",
        "[text(1,3):Foo\nbar\nbaz::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>Foo
bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_206c():
    """
    Test case 206:  Here is a simple example:
    """

    # Arrange
    source_markdown = """> Foo
 bar
  baz
   bofo"""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n\n]",
        "[para(1,3):\n \n  \n   ]",
        "[text(1,3):Foo\nbar\nbaz\nbofo::\n\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>Foo
bar
baz
bofo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_207():
    """
    Test case 207:  The spaces after the > characters can be omitted:
    """

    # Arrange
    source_markdown = """># Foo
>bar
> baz"""
    expected_tokens = [
        "[block-quote(1,1)::>\n>\n> ]",
        "[atx(1,2):1:0:]",
        "[text(1,4):Foo: ]",
        "[end-atx::]",
        "[para(2,2):\n]",
        "[text(2,2):bar\nbaz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h1>Foo</h1>
<p>bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_208():
    """
    Test case 208:  (part 1) The > characters can be indented 1-3 spaces:
    """

    # Arrange
    source_markdown = """   > # Foo
   > bar
 > baz"""
    expected_tokens = [
        "[block-quote(1,4):   :   > \n   > \n > ]",
        "[atx(1,6):1:0:]",
        "[text(1,8):Foo: ]",
        "[end-atx::]",
        "[para(2,6):\n]",
        "[text(2,6):bar\nbaz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h1>Foo</h1>
<p>bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_209():
    """
    Test case 209:  Four spaces gives us a code block:
    """

    # Arrange
    source_markdown = """    > # Foo
    > bar
    > baz"""
    expected_tokens = [
        "[icode-block(1,5):    :\n    \n    ]",
        "[text(1,5):\a>\a&gt;\a # Foo\n\a>\a&gt;\a bar\n\a>\a&gt;\a baz:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<pre><code>&gt; # Foo
&gt; bar
&gt; baz
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_210():
    """
    Test case 210:  The Laziness clause allows us to omit the > before paragraph continuation text:
    """

    # Arrange
    source_markdown = """> # Foo
> bar
baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n]",
        "[atx(1,3):1:0:]",
        "[text(1,5):Foo: ]",
        "[end-atx::]",
        "[para(2,3):\n]",
        "[text(2,3):bar\nbaz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h1>Foo</h1>
<p>bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_211():
    """
    Test case 211:  A block quote can contain some lazy and some non-lazy continuation lines:
    """

    # Arrange
    source_markdown = """> bar
baz
> foo"""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n> ]",
        "[para(1,3):\n\n]",
        "[text(1,3):bar\nbaz\nfoo::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar
baz
foo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_211a():
    """
    Test case 211a:  variations
    """

    # Arrange
    source_markdown = """> bar
> baz
> foo"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):\n\n]",
        "[text(1,3):bar\nbaz\nfoo::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar
baz
foo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_211b():
    """
    Test case 211b:  variations
    """

    # Arrange
    source_markdown = """> bar

> baz

> foo"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> \n]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
        "[block-quote(5,1)::> ]",
        "[para(5,3):]",
        "[text(5,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar</p>
</blockquote>
<blockquote>
<p>baz</p>
</blockquote>
<blockquote>
<p>foo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_211c():
    """
    Test case 211c:  variations
    """

    # Arrange
    source_markdown = """> bar
>
> baz
>
> foo"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):bar:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[text(5,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar</p>
<p>baz</p>
<p>foo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_212():
    """
    Test case 212:  Laziness only applies to lines that would have been continuations of paragraphs had they been prepended with block quote markers.
    """

    # Arrange
    source_markdown = """> foo
---"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(2,1):-::---]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
</blockquote>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_212a():
    """
    Test case 212a:  variation
    """

    # Arrange
    source_markdown = """> foo
> ---"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[setext(2,3):-:3::(1,3)]",
        "[text(1,3):foo:]",
        "[end-setext::]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h2>foo</h2>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_212b():
    """
    Test case 212b:  variation
    """

    # Arrange
    source_markdown = """> foo

> ---"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[tbreak(3,3):-::---]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
</blockquote>
<blockquote>
<hr />
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_212c():
    """
    Test case 212c:  variation
    """

    # Arrange
    source_markdown = """> foo
>
> ---"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[tbreak(3,3):-::---]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<hr />
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213():
    """
    Test case 213:  then the block quote ends after the first line:
    """

    # Arrange
    source_markdown = """> - foo
- bar"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
        "[ulist(2,1):-::2:]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo</li>
</ul>
</blockquote>
<ul>
<li>bar</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213ax():
    """
    Test case 213a:  variation
    """

    # Arrange
    source_markdown = """> - foo
> - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):foo:]",
        "[end-para:::True]",
        "[li(2,3):4:  :]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo</li>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213aa():
    """
    Test case 213aa:  variation
    """

    # Arrange
    source_markdown = """> - foo
>   - boo
> - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):foo:]",
        "[end-para:::True]",
        "[ulist(2,5):-::6:    ]",
        "[para(2,7):]",
        "[text(2,7):boo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo
<ul>
<li>boo</li>
</ul>
</li>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213ab():
    """
    Test case 213ab:  variation
    """

    # Arrange
    source_markdown = """- foo
  - boo
- bar"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):boo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(3,1):2::]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>boo</li>
</ul>
</li>
<li>bar</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213b():
    """
    Test case 213b:  variation
    """

    # Arrange
    source_markdown = """> - foo
> - bar
> - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):foo:]",
        "[end-para:::True]",
        "[li(2,3):4:  :]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo</li>
<li>bar</li>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213c():
    """
    Test case 213b:  variation
    """

    # Arrange
    source_markdown = """> - foo
    brr
> - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[ulist(1,3):-::4:  :    ]",
        "[para(1,5):\n]",
        "[text(1,5):foo\nbrr::\n]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo
brr</li>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_213d():
    """
    Test case 213d:  variation
    """

    # Arrange
    source_markdown = """> - foo
>   brr
> - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):\n  ]",
        "[text(1,5):foo\nbrr::\n]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>foo
brr</li>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_214():
    """
    Test case 214:  (part 1) For the same reason, we can’t omit the > in front of subsequent lines of an indented or fenced code block:
    Variation: - test_paragraph_series_n_bq_i4_t_nl_bq_i4_t
                 - proper indent and block quotes on both
               - test_paragraph_series_n_bq_i4_t_nl_bq_i3_t
                 - same as above, but second line with 1 less ws
               - test_paragraph_series_n_bq_i4_t_nl_bq_nl_bq_i4_t
                 - over multiple lines, each starting with block quote
               - test_paragraph_series_n_bq_i4_t_nl_nl_bq_i4_t
                 - over multiple lines, blank lines without block quote
    """

    # Arrange
    source_markdown = """>     foo
    bar"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[icode-block(1,7):    :]",
        "[text(1,7):foo:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[icode-block(2,5):    :]",
        "[text(2,5):bar:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>foo
</code></pre>
</blockquote>
<pre><code>bar
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_215():
    """
    Test case 215:  (part 2) For the same reason, we can’t omit the > in front of subsequent lines of an indented or fenced code block:

    Variation: - test_paragraph_series_n_bq_t_nl_bq_fb
                 - for the entire block to be considered a fcb, all must be prefaced
                   with ">"
               - test_paragraph_series_n_bq_t_nl_bq_fb_nl_with_bq
                 - for any blank lines within block to not abort block quotes, must
                   prefix ">"
               - test_paragraph_series_n_bq_t_nl_bq_fb_nl_without_bq
                 - same as above, but lack of ">" causes block to be split up
    """

    # Arrange
    source_markdown = """> ```
foo
```"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[fcode-block(1,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[para(2,1):]",
        "[text(2,1):foo:]",
        "[end-para:::False]",
        "[fcode-block(3,1):`:3::::::]",
        "[end-fcode-block:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code></code></pre>
</blockquote>
<p>foo</p>
<pre><code></code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_216():
    """
    Test case 216:  Note that in the following case, we have a lazy continuation line:
    """

    # Arrange
    source_markdown = """> foo
    - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n    ]",
        "[text(1,3):foo\n- bar::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo
- bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_216a():
    """
    Test case 216a:  variation
    """

    # Arrange
    source_markdown = """> foo
>    - bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,6):-::7:     ]",
        "[para(2,8):]",
        "[text(2,8):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<ul>
<li>bar</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_217():
    """
    Test case 217:  (part 1) A block quote can be empty:
    """

    # Arrange
    source_markdown = """>"""
    expected_tokens = [
        "[block-quote(1,1)::>]",
        "[BLANK(1,2):]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_218():
    """
    Test case 218:  (part 2) A block quote can be empty:
    """

    # Arrange
    source_markdown = """>
>\a\a
> """.replace(
        "\a", " "
    )
    expected_tokens = [
        "[block-quote(1,1)::>\n> \n> ]",
        "[BLANK(1,2):]",
        "[BLANK(2,3): ]",
        "[BLANK(3,3):]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_219():
    """
    Test case 219:  A block quote can have initial or final blank lines:
    """

    # Arrange
    source_markdown = """>
> foo
>  """
    expected_tokens = [
        "[block-quote(1,1)::>\n> \n> ]",
        "[BLANK(1,2):]",
        "[para(2,3):]",
        "[text(2,3):foo:]",
        "[end-para:::True]",
        "[BLANK(3,3): ]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_220():
    """
    Test case 220:  A blank line always separates block quotes:
    """

    # Arrange
    source_markdown = """> foo

> bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
</blockquote>
<blockquote>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_221():
    """
    Test case 221:  Consecutiveness means that if we put these block quotes together, we get a single block quote:
    """

    # Arrange
    source_markdown = """> foo
> bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):\n]",
        "[text(1,3):foo\nbar::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo
bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_222():
    """
    Test case 222:  To get a block quote with two paragraphs, use:
    """

    # Arrange
    source_markdown = """> foo
>
> bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_222a():
    """
    Test case 222a:  To get a block quote with two paragraphs, use:
    """

    # Arrange
    source_markdown = """> foo
>\a
> bar""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,3):]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_222b():
    """
    Test case 222a:  To get a block quote with two paragraphs, use:
    """

    # Arrange
    source_markdown = """> foo
>\a\a
> bar""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,3): ]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_222c():
    """
    Test case 222a:  To get a block quote with two paragraphs, use:
    """

    # Arrange
    source_markdown = """> foo
>
>
> bar"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_223():
    """
    Test case 223:  Block quotes can interrupt paragraphs:
    """

    # Arrange
    source_markdown = """foo
> bar"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[end-para:::True]",
        "[block-quote(2,1)::> ]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<p>foo</p>
<blockquote>
<p>bar</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_224():
    """
    Test case 224:  In general, blank lines are not needed before or after block quotes:
    """

    # Arrange
    source_markdown = """> aaa
***
> bbb"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):aaa:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(2,1):*::***]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):bbb:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>aaa</p>
</blockquote>
<hr />
<blockquote>
<p>bbb</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_225():
    """
    Test case 225:  (part 1) However, because of laziness, a blank line is needed between a block quote and a following paragraph:
    """

    # Arrange
    source_markdown = """> bar
baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n]",
        "[text(1,3):bar\nbaz::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar
baz</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_226():
    """
    Test case 226:  (part 2) However, because of laziness, a blank line is needed between a block quote and a following paragraph:
    """

    # Arrange
    source_markdown = """> bar

baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar</p>
</blockquote>
<p>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_227():
    """
    Test case 227:  (part 3) However, because of laziness, a blank line is needed between a block quote and a following paragraph:
    """

    # Arrange
    source_markdown = """> bar
>
baz"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n]",
        "[para(1,3):]",
        "[text(1,3):bar:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        "[end-block-quote:::False]",
        "[para(3,1):]",
        "[text(3,1):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<blockquote>
<p>bar</p>
</blockquote>
<p>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_228():
    """
    Test case 228:  (part 1) It is a consequence of the Laziness rule that any number of initial >s may be omitted on a continuation line of a nested block quote:
    """

    # Arrange
    source_markdown = """> > > foo
bar"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,3)::]",
        "[block-quote(1,5)::> > > \n]",
        "[para(1,7):\n]",
        "[text(1,7):foo\nbar::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<p>foo
bar</p>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229():
    """
    Test case 229:  (part 2) It is a consequence of the Laziness rule that any number of initial >s may be omitted on a continuation line of a nested block quote:
    """

    # Arrange
    source_markdown = """>>> foo
> bar
>>baz"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> \n> \n>>]",
        "[para(1,5):\n\n]",
        "[text(1,5):foo\nbar\nbaz::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<p>foo
bar
baz</p>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229a():
    """
    Test case 229a:  variation
    """

    # Arrange
    source_markdown = """> 1
>> 2
> 1
>> > 3
> > 2
"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):1:]",
        "[end-para:::True]",
        "[block-quote(2,1)::>> \n> ]",
        "[para(2,4):\n]",
        "[text(2,4):2\n1::\n]",
        "[end-para:::True]",
        "[block-quote(4,1)::>> > \n> > \n]",
        "[para(4,6):\n]",
        "[text(4,6):3\n2::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(6,1):]",
    ]
    expected_gfm = """<blockquote>
<p>1</p>
<blockquote>
<p>2
1</p>
<blockquote>
<p>3
2</p>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229b():
    """
    Test case 229b:  variation
    """

    # Arrange
    source_markdown = """> 1

>> 2

> 1

>> > 3

> > 2
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):1:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::]",
        "[block-quote(3,2)::>> \n]",
        "[para(3,4):]",
        "[text(3,4):2:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
        "[block-quote(5,1)::> \n]",
        "[para(5,3):]",
        "[text(5,3):1:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(6,1):]",
        "[block-quote(7,1)::]",
        "[block-quote(7,2)::]",
        "[block-quote(7,4)::>> > \n]",
        "[para(7,6):]",
        "[text(7,6):3:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(8,1):]",
        "[block-quote(9,1)::]",
        "[block-quote(9,3)::> > \n]",
        "[para(9,5):]",
        "[text(9,5):2:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(10,1):]",
    ]
    expected_gfm = """<blockquote>
<p>1</p>
</blockquote>
<blockquote>
<blockquote>
<p>2</p>
</blockquote>
</blockquote>
<blockquote>
<p>1</p>
</blockquote>
<blockquote>
<blockquote>
<blockquote>
<p>3</p>
</blockquote>
</blockquote>
</blockquote>
<blockquote>
<blockquote>
<p>2</p>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229cx():
    """
    Test case 229b:  variation
    """

    # Arrange
    source_markdown = """>     1
>>     2
"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[icode-block(1,7):    :]",
        "[text(1,7):1:]",
        "[end-icode-block:::True]",
        "[block-quote(2,1)::>> \n]",
        "[icode-block(2,8):    :]",
        "[text(2,8):2:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>1
</code></pre>
<blockquote>
<pre><code>2
</code></pre>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ca():
    """
    Test case 229b:  variation
    """

    # Arrange
    source_markdown = """>     1
>>>     2
"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[icode-block(1,7):    :]",
        "[text(1,7):1:]",
        "[end-icode-block:::True]",
        "[block-quote(2,1)::]",
        "[block-quote(2,2)::>>> \n]",
        "[icode-block(2,9):    :]",
        "[text(2,9):2:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>1
</code></pre>
<blockquote>
<blockquote>
<pre><code>2
</code></pre>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229dx():
    """
    Test case 229d:  variation
    """

    # Arrange
    source_markdown = """>>     1
>     2
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::>> ]",
        "[icode-block(1,8):    :]",
        "[text(1,8):1:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[icode-block(2,7):    :]",
        "[text(2,7):2:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<pre><code>1
</code></pre>
</blockquote>
<pre><code>2
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229da():
    """
    Test case 229d:  variation
    """

    # Arrange
    source_markdown = """>>>     1
>     2
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> ]",
        "[icode-block(1,9):    :]",
        "[text(1,9):1:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[icode-block(2,7):    :]",
        "[text(2,7):2:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(3,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<pre><code>1
</code></pre>
</blockquote>
</blockquote>
<pre><code>2
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ex():
    """
    Test case 229d:  variation
    """

    # Arrange
    source_markdown = """> ```
>> 2
>> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n>\n]",
        "[fcode-block(1,3):`:3::::::]",
        "[text(2,2):\a>\a&gt;\a 2\n\a>\a&gt;\a ```:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>&gt; 2
&gt; ```
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ea():
    """
    Test case 229d:  variation
    """

    # Arrange
    source_markdown = """> ```
>>> 2
>>> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n>\n]",
        "[fcode-block(1,3):`:3::::::]",
        "[text(2,2):\a>\a&gt;\a\a>\a&gt;\a 2\n\a>\a&gt;\a\a>\a&gt;\a ```:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>&gt;&gt; 2
&gt;&gt; ```
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229fx():
    """
    Test case 229e:  variation
    """

    # Arrange
    source_markdown = """> ```
> 2
>> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n]",
        "[fcode-block(1,3):`:3::::::]",
        "[text(2,3):2\n\a>\a&gt;\a ```:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>2
&gt; ```
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229fa():
    """
    Test case 229e:  variation
    """

    # Arrange
    source_markdown = """> ```
> 2
>>> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n]",
        "[fcode-block(1,3):`:3::::::]",
        "[text(2,3):2\n\a>\a&gt;\a\a>\a&gt;\a ```:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<pre><code>2
&gt;&gt; ```
</code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229gx():
    """
    Test case 229f:  variation
    """

    # Arrange
    source_markdown = """>> ```
> 2
> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n]",
        "[block-quote(1,2)::>> ]",
        "[fcode-block(1,4):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[para(2,3):]",
        "[text(2,3):2:]",
        "[end-para:::False]",
        "[fcode-block(3,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<pre><code></code></pre>
</blockquote>
<p>2</p>
<pre><code></code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ga():
    """
    Test case 229f:  variation
    """

    # Arrange
    source_markdown = """>>> ```
> 2
> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> ]",
        "[fcode-block(1,5):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[para(2,3):]",
        "[text(2,3):2:]",
        "[end-para:::False]",
        "[fcode-block(3,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<pre><code></code></pre>
</blockquote>
</blockquote>
<p>2</p>
<pre><code></code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229hx():
    """
    Test case 229g:  variation
    """

    # Arrange
    source_markdown = """>> ```
>> 2
> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::>> \n>> ]",
        "[fcode-block(1,4):`:3::::::]",
        "[text(2,4):2:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<pre><code>2
</code></pre>
</blockquote>
<pre><code></code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ha():
    """
    Test case 229g:  variation
    """

    # Arrange
    source_markdown = """>>> ```
>>> 2
> ```
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> \n>>> ]",
        "[fcode-block(1,5):`:3::::::]",
        "[text(2,5):2:]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,3):`:3::::::]",
        "[end-fcode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<pre><code>2
</code></pre>
</blockquote>
</blockquote>
<pre><code></code></pre>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ix():
    """
    Test case 229g:  variation
    """

    # Arrange
    source_markdown = """> <script>
> comments
>> </script>
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n]",
        "[html-block(1,3)]",
        "[text(1,3):<script>\ncomments\n> </script>:]",
        "[end-html-block:::False]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<script>
comments
> </script>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ia():
    """
    Test case 229ia:  variation
    """

    # Arrange
    source_markdown = """> <script>
> comments
>>> </script>
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n]",
        "[html-block(1,3)]",
        "[text(1,3):<script>\ncomments\n>> </script>:]",
        "[end-html-block:::False]",
        "[end-block-quote:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<blockquote>
<script>
comments
>> </script>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229jx():
    """
    Test case 229j:  variation
    """

    # Arrange
    source_markdown = """>> <script>
>> comments
> </script>
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::>> \n>> ]",
        "[html-block(1,4)]",
        "[text(1,4):<script>\ncomments:]",
        "[end-html-block:::True]",
        "[end-block-quote:::True]",
        "[html-block(3,3)]",
        "[text(3,3):</script>:]",
        "[end-html-block:::False]",
        "[BLANK(4,1):]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<script>
comments
</blockquote>
</script>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_229ja():
    """
    Test case 229j:  variation
    """

    # Arrange
    source_markdown = """>>> <script>
>>> comments
> </script>
"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> \n>>> ]",
        "[html-block(1,5)]",
        "[text(1,5):<script>\ncomments:]",
        "[end-html-block:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[html-block(3,3)]",
        "[text(3,3):</script>:]",
        "[end-html-block:::False]",
        "[BLANK(4,1):]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<script>
comments
</blockquote>
</blockquote>
</script>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_230():
    """
    Test case 230:  When including an indented code block in a block quote, remember that the block quote marker includes both the > and a following space. So five spaces are needed after the >:
    """

    # Arrange
    source_markdown = """>     code

>    not code"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[icode-block(1,7):    :]",
        "[text(1,7):code:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[para(3,6):   ]",
        "[text(3,6):not code:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>code
</code></pre>
</blockquote>
<blockquote>
<p>not code</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_01():
    """
    Test case Bq01:  Indents should work properly for a list block containing a
                     block quote where the block quote ends and there is
                     more data for that item within the list
    """

    # Arrange
    source_markdown = """* start
  > quote
* end"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(3,1):2::]",
        "[para(3,3):]",
        "[text(3,3):end:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>start
<blockquote>
<p>quote</p>
</blockquote>
</li>
<li>end</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_01a():
    """
    Test case Bq01a:  same as 01, just more levels of list
    """

    # Arrange
    source_markdown = """* start
  > quote
  * middle
    > middle
      quote
  * more middle
* end"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[ulist(3,3):*::4:  :    ]",
        "[para(3,5):]",
        "[text(3,5):middle:]",
        "[end-para:::True]",
        "[block-quote(4,5):    :    > \n    ]",
        "[para(4,7):\n  ]",
        "[text(4,7):middle\nquote::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(6,3):4:  :]",
        "[para(6,5):]",
        "[text(6,5):more middle:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):end:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>start
<blockquote>
<p>quote</p>
</blockquote>
<ul>
<li>middle
<blockquote>
<p>middle
quote</p>
</blockquote>
</li>
<li>more middle</li>
</ul>
</li>
<li>end</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02x():
    """
    Test case Bq02:  Indents should work properly for a block quote containing a
                     list block where the list block ends and there is
                     more data for that block quote within the list
    """

    # Arrange
    source_markdown = """> start
> - quote
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):\n]",
        "[text(2,5):quote\nend::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start</p>
<ul>
<li>quote
end</li>
</ul>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02ax():
    """
    Test case Bq02a:  variant
    """

    # Arrange
    source_markdown = """> start
> - quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start</p>
<ul>
<li>quote</li>
</ul>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02aa():
    """
    Test case Bq02aa:  variant
    """

    # Arrange
    source_markdown = """- quote

end"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):quote:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[end-ulist:::True]",
        "[para(3,1):]",
        "[text(3,1):end:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<ul>
<li>quote</li>
</ul>
<p>end</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02ab():
    """
    Test case Bq02ab:  variant
    """

    # Arrange
    source_markdown = """> - start
> quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):\n]",
        "[text(1,5):start\nquote::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>start
quote</li>
</ul>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02ac():
    """
    Test case Bq02ac:  variant
    """

    # Arrange
    source_markdown = """> - start
> - quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):start:]",
        "[end-para:::True]",
        "[li(2,3):4:  :]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>start</li>
<li>quote</li>
</ul>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_02ad():
    """
    Test case Bq02ad:  variant
    """

    # Arrange
    source_markdown = """> - start
>   - quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[ulist(1,3):-::4:  ]",
        "[para(1,5):]",
        "[text(1,5):start:]",
        "[end-para:::True]",
        "[ulist(2,5):-::6:    ]",
        "[para(2,7):]",
        "[text(2,7):quote:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<ul>
<li>start
<ul>
<li>quote</li>
</ul>
</li>
</ul>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03x():
    """
    Test case Bq03:  link definition within a block quote
                     copy of test_link_reference_definitions_161 but within single block
    """

    # Arrange
    source_markdown = """> [foo]: /url "title"
>
> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> ]",
        '[link-ref-def(1,3):True::foo:: :/url:: :title:"title":]',
        "[BLANK(2,2):]",
        "[para(3,3):]",
        "[link(3,3):shortcut:/url:title::::foo:::::]",
        "[text(3,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="/url" title="title">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03xa():
    """
    Test case Bq03:  link definition within a block quote
                     copy of test_link_reference_definitions_161 but within single block
    """

    # Arrange
    source_markdown = """>
> [foo]: /url "title"
>
> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::>\n> \n>\n> ]",
        "[BLANK(1,2):]",
        '[link-ref-def(2,3):True::foo:: :/url:: :title:"title":]',
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[link(4,3):shortcut:/url:title::::foo:::::]",
        "[text(4,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="/url" title="title">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03xb():
    """
    Test case Bq03:  link definition within a block quote
                     copy of test_link_reference_definitions_161 but within single block
    """

    # Arrange
    source_markdown = """> abc
>
> [foo]: /url "title"
>
> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n>\n> \n>\n> ]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[BLANK(2,2):]",
        '[link-ref-def(3,3):True::foo:: :/url:: :title:"title":]',
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[link(5,3):shortcut:/url:title::::foo:::::]",
        "[text(5,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>abc</p>
<p><a href="/url" title="title">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03a():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_161 but within
                      two distinct blocks
    """

    # Arrange
    source_markdown = """> [foo]: /url "title"

> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        '[link-ref-def(1,3):True::foo:: :/url:: :title:"title":]',
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[link(3,3):shortcut:/url:title::::foo:::::]",
        "[text(3,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
</blockquote>
<blockquote>
<p><a href="/url" title="title">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03aa():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_161 but within
                      two distinct blocks
    """

    # Arrange
    source_markdown = """> [foo]: /url "title"
> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        '[link-ref-def(1,3):True::foo:: :/url:: :title:"title":]',
        "[para(2,3):]",
        "[link(2,3):shortcut:/url:title::::foo:::::]",
        "[text(2,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="/url" title="title">foo</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03ab():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_161 but within
                      two distinct blocks
    """

    # Arrange
    source_markdown = """> abc

> [foo]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):]",
        "[text(1,3):abc:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[BLANK(2,1):]",
        "[block-quote(3,1)::> ]",
        "[para(3,3):]",
        "[text(3,3):[:]",
        "[text(3,4):foo:]",
        "[text(3,7):]:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>abc</p>
</blockquote>
<blockquote>
<p>[foo]</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03b():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_164 but within
                      a single block
    """

    # Arrange
    source_markdown = """> [Foo bar]:
> <my url>
> 'title'
>
> [Foo bar]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n>\n> ]",
        "[link-ref-def(1,3):True::foo bar:Foo bar:\n:my%20url:<my url>:\n:title:'title':]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[link(5,3):shortcut:my%20url:title::::Foo bar:::::]",
        "[text(5,4):Foo bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="my%20url" title="title">Foo bar</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03ba():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_164 but within
                      a single block
    """

    # Arrange
    source_markdown = """> [Foo bar]:
> <my url>
> 'my
> title'
>
> [Foo bar]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n> \n>\n> ]",
        "[link-ref-def(1,3):True::foo bar:Foo bar:\n:my%20url:<my url>:\n:my\ntitle:'my\ntitle':]",
        "[BLANK(5,2):]",
        "[para(6,3):]",
        "[link(6,3):shortcut:my%20url:my\ntitle::::Foo bar:::::]",
        "[text(6,4):Foo bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="my%20url" title="my
title">Foo bar</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_03bb():
    """
    Test case Bq03a:  link definition within a block quote
                      copy of test_link_reference_definitions_164 but within
                      a single block
    """

    # Arrange
    source_markdown = """> [Foo
> bar]:
> <my url>
> 'my
> title'
>
> [Foo
> bar]"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n> \n> \n>\n> \n> ]",
        "[link-ref-def(1,3):True::foo bar:Foo\nbar:\n:my%20url:<my url>:\n:my\ntitle:'my\ntitle':]",
        "[BLANK(6,2):]",
        "[para(7,3):\n]",
        "[link(7,3):shortcut:my%20url:my\ntitle::::Foo\nbar:::::]",
        "[text(7,4):Foo\nbar::\n]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="my%20url" title="my
title">Foo
bar</a></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> start
> quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[para(1,3):\n]",
        "[text(1,3):start\nquote::\n]",
        "[end-para:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start
quote</p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04a():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> start
> *the* quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[para(1,3):\n]",
        "[text(1,3):start\n::\n]",
        "[emphasis(2,1):1:*]",
        "[text(2,2):the:]",
        "[end-emphasis(2,5)::]",
        "[text(2,6): quote:]",
        "[end-para:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start
<em>the</em> quote</p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04b():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> start
> *the* quote
> ---
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n>\n> ]",
        "[setext(3,3):-:3::(1,3)]",
        "[text(1,3):start\n::\n]",
        "[emphasis(2,1):1:*]",
        "[text(2,2):the:]",
        "[end-emphasis(2,5)::]",
        "[text(2,6): quote:]",
        "[end-setext::]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[text(5,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<h2>start
<em>the</em> quote</h2>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04c():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> ```
> start
> *the* quote
> ```
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n> \n>\n> ]",
        "[fcode-block(1,3):`:3::::::]",
        "[text(2,3):start\n*the* quote:]",
        "[end-fcode-block::3:False]",
        "[BLANK(5,2):]",
        "[para(6,3):]",
        "[text(6,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>start
*the* quote
</code></pre>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04d():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """>     start
>     *the* quote
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[icode-block(1,7):    :\n    ]",
        "[text(1,7):start\n*the* quote:]",
        "[end-icode-block:::False]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<pre><code>start
*the* quote
</code></pre>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04e():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> <!--
> script
> -->
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n>\n> ]",
        "[html-block(1,3)]",
        "[text(1,3):<!--\nscript\n-->:]",
        "[end-html-block:::False]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[text(5,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<!--
script
-->
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04f():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> [
> abc
> ](/uri)
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n>\n> ]",
        "[para(1,3):\n\n]",
        "[link(1,3):inline:/uri:::::\nabc\n:False::::]",
        "[text(1,4):\nabc\n::\n\n]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[text(5,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a href="/uri">
abc
</a></p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04g():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> ![
> abc
> ](/uri)
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> \n>\n> ]",
        "[para(1,3):\n\n]",
        "[image(1,3):inline:/uri::\nabc\n::::\nabc\n:False::::]",
        "[end-para:::True]",
        "[BLANK(4,2):]",
        "[para(5,3):]",
        "[text(5,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><img src="/uri" alt="
abc
" /></p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04h():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> a <html
> is="1">
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[para(1,3):\n]",
        "[text(1,3):a :]",
        '[raw-html(1,5):html\nis="1"]',
        "[end-para:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>a <html
is="1"></p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_04j():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> a``html
> maybe``
>
> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n>\n> ]",
        "[para(1,3):\n]",
        "[text(1,3):a:]",
        "[icode-span(1,4):html\a\n\a \amaybe:``::]",
        "[end-para:::True]",
        "[BLANK(3,2):]",
        "[para(4,3):]",
        "[text(4,3):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>a<code>html maybe</code></p>
<p>end</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_05x():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """>> start
>> - quote
> end"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,2)::>> \n>> \n> ]",
        "[para(1,4):]",
        "[text(1,4):start:]",
        "[end-para:::True]",
        "[ulist(2,4):-::5:   ]",
        "[para(2,6):\n]",
        "[text(2,6):quote\nend::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<p>start</p>
<ul>
<li>quote
end</li>
</ul>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_05xa():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """>>> start
>>> - quote
> end"""
    expected_tokens = [
        "[block-quote(1,1)::]",
        "[block-quote(1,2)::]",
        "[block-quote(1,3)::>>> \n>>> \n> ]",
        "[para(1,5):]",
        "[text(1,5):start:]",
        "[end-para:::True]",
        "[ulist(2,5):-::6:    ]",
        "[para(2,7):\n]",
        "[text(2,7):quote\nend::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<blockquote>
<blockquote>
<p>start</p>
<ul>
<li>quote
end</li>
</ul>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_05a():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> start
> - quote
>> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::>> ]",
        "[para(3,4):]",
        "[text(3,4):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start</p>
<ul>
<li>quote</li>
</ul>
<blockquote>
<p>end</p>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_block_quotes_extra_05aa():
    """
    Test case Bq04:  variant
    """

    # Arrange
    source_markdown = """> start
> - quote
>>> end"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):]",
        "[text(1,3):start:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):quote:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[block-quote(3,1)::]",
        "[block-quote(3,2)::>>> ]",
        "[para(3,5):]",
        "[text(3,5):end:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>start</p>
<ul>
<li>quote</li>
</ul>
<blockquote>
<blockquote>
<p>end</p>
</blockquote>
</blockquote>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide processing for the list blocks.
"""

import logging
import string

from pymarkdown.container_markdown_token import (
    NewListItemMarkdownToken,
    OrderedListStartMarkdownToken,
    UnorderedListStartMarkdownToken,
)
from pymarkdown.leaf_block_processor import LeafBlockProcessor
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.stack_token import OrderedListStackToken, UnorderedListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-many-lines
class ListBlockProcessor:
    """
    Class to provide processing for the list blocks.
    """

    __ulist_start_characters = "-+*"
    __olist_start_characters = ".)"

    @staticmethod
    # pylint: disable=too-many-arguments
    def is_ulist_start(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        skip_whitespace_check,
        adj_ws=None,
    ):
        """
        Determine if we have the start of an un-numbered list.
        """
        POGGER.debug("is_ulist_start>>pre>>")
        is_start, after_all_whitespace_index = False, -1
        if adj_ws is None:
            adj_ws = extracted_whitespace

        if (
            ParserHelper.is_length_less_than_or_equal_to(adj_ws, 3)
            or skip_whitespace_check
        ):
            is_start = ListBlockProcessor.__is_start_ulist(
                line_to_parse, start_index, extracted_whitespace
            )
        if is_start:
            (
                is_start,
                after_all_whitespace_index,
            ) = ListBlockProcessor.__is_start_phase_one(
                parser_state, line_to_parse, start_index, False
            )
        if is_start:
            is_start = ListBlockProcessor.__is_start_phase_two(
                parser_state,
                line_to_parse[start_index],
                True,
                False,
                after_all_whitespace_index,
                line_to_parse,
                start_index,
            )

        return is_start, after_all_whitespace_index, start_index, 0
        # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
    @staticmethod
    def is_olist_start(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        skip_whitespace_check,
        adj_ws=None,
    ):
        """
        Determine if we have the start of an numbered or ordered list.
        """

        is_start, after_all_whitespace_index, index, number_of_digits, is_not_one = (
            False,
            -1,
            None,
            None,
            False,
        )
        if adj_ws is None:
            adj_ws = extracted_whitespace

        if (
            ParserHelper.is_length_less_than_or_equal_to(adj_ws, 3)
            or skip_whitespace_check
        ):
            (
                is_start,
                index,
                number_of_digits,
                is_not_one,
            ) = ListBlockProcessor.__is_start_olist(line_to_parse, start_index)
        if is_start:
            (
                is_start,
                after_all_whitespace_index,
            ) = ListBlockProcessor.__is_start_phase_one(
                parser_state, line_to_parse, index, is_not_one
            )
        if is_start:
            is_start = ListBlockProcessor.__is_start_phase_two(
                parser_state,
                line_to_parse[index],
                False,
                is_not_one,
                after_all_whitespace_index,
                line_to_parse,
                start_index,
            )

        return is_start, after_all_whitespace_index, index, number_of_digits

    # pylint: enable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements

    @staticmethod
    def __is_start_ulist(line_to_parse, start_index, extracted_whitespace):
        is_start = ParserHelper.is_character_at_index_one_of(
            line_to_parse, start_index, ListBlockProcessor.__ulist_start_characters
        )

        # Thematic breaks have precedence, so stop a list start if we find one.
        if is_start:
            is_break, _ = LeafBlockProcessor.is_thematic_break(
                line_to_parse, start_index, extracted_whitespace
            )
            is_start = is_start and not is_break
        return is_start

    @staticmethod
    def __is_start_olist(line_to_parse, start_index):
        index, number_of_digits, is_not_one = None, None, None
        is_start = ParserHelper.is_character_at_index_one_of(
            line_to_parse, start_index, string.digits
        )
        if is_start:
            index = start_index
            while ParserHelper.is_character_at_index_one_of(
                line_to_parse, index, string.digits
            ):
                index += 1
            olist_index_number, number_of_digits = (
                line_to_parse[start_index:index],
                index - start_index,
            )
            POGGER.debug("olist?$<<count>>$<<", olist_index_number, number_of_digits)
            is_not_one = olist_index_number != "1"

            is_olist_start = ParserHelper.is_character_at_index_one_of(
                line_to_parse, index, ListBlockProcessor.__olist_start_characters
            )
            POGGER.debug("is_olist_start>>$", is_olist_start)
            is_start = is_olist_start and number_of_digits <= 9

        return is_start, index, number_of_digits, is_not_one

    @staticmethod
    def __is_start_phase_one(parser_state, line_to_parse, start_index, is_not_one):

        is_start, line_to_parse_size = False, len(line_to_parse)
        after_all_whitespace_index, _ = ParserHelper.extract_whitespace(
            line_to_parse, start_index + 1
        )
        POGGER.debug(
            "after_all_whitespace_index>>$>>len>>$",
            after_all_whitespace_index,
            line_to_parse_size,
        )
        at_end_of_line = after_all_whitespace_index == line_to_parse_size
        POGGER.debug("at_end_of_line>>$", at_end_of_line)

        is_in_paragraph = parser_state.token_stack[-1].is_paragraph
        is_paragraph_in_list = (
            parser_state.token_stack[-2].is_list if is_in_paragraph else False
        )

        is_start = not (
            is_in_paragraph
            and not is_paragraph_in_list
            and (at_end_of_line or is_not_one)
        ) and (
            ParserHelper.is_character_at_index_whitespace(
                line_to_parse, start_index + 1
            )
            or ((start_index + 1) == line_to_parse_size)
        )
        return is_start, after_all_whitespace_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __is_start_phase_two(
        parser_state,
        xx_seq,
        is_unordered_list,
        is_not_one,
        after_all_whitespace_index,
        line_to_parse,
        start_index,
    ):
        (
            is_start,
            is_first_item_in_list,
            is_sub_list,
            is_in_paragraph,
            at_end_of_line,
        ) = (
            True,
            False,
            False,
            parser_state.token_stack[-1].is_paragraph,
            (after_all_whitespace_index == len(line_to_parse)),
        )
        POGGER.debug("is_in_paragraph>>$", is_in_paragraph)
        POGGER.debug("at_end_of_line>>$", at_end_of_line)

        if is_in_paragraph:
            if not parser_state.token_stack[-2].is_list:
                POGGER.debug(
                    "top of stack is not list>>$", parser_state.token_stack[-2]
                )
                is_first_item_in_list = True
            elif is_unordered_list and parser_state.token_stack[-2].is_ordered_list:
                POGGER.debug(
                    "top of stack is ordered list>>$", parser_state.token_stack[-2]
                )
                is_first_item_in_list = True
            elif xx_seq != parser_state.token_stack[-2].list_character[-1]:
                POGGER.debug(
                    "xx>>$!=$",
                    line_to_parse[start_index],
                    parser_state.token_stack[-2].list_character,
                )
                is_first_item_in_list = True
            else:
                is_first_item_in_list = (
                    start_index >= parser_state.token_stack[-2].indent_level
                )
                POGGER.debug(
                    "start_index>>$>=$",
                    start_index,
                    parser_state.token_stack[-2].indent_level,
                )
            POGGER.debug("is_first_item_in_list>>$", is_first_item_in_list)

            if parser_state.token_stack[-2].is_list:
                POGGER.debug("old_indent=$", parser_state.token_stack[-2].indent_level)
                POGGER.debug("new_indent=$", start_index)
                is_sub_list = start_index >= parser_state.token_stack[-2].indent_level

        POGGER.debug(
            "is_in_para>>$>>EOL>$>is_first>$",
            is_in_paragraph,
            at_end_of_line,
            is_first_item_in_list,
        )
        if (
            is_in_paragraph
            and (at_end_of_line or is_not_one)
            and is_first_item_in_list
            and is_sub_list
        ):
            is_start = False
            POGGER.debug("is_start>>$", is_start)
        return is_start

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-locals, too-many-arguments
    @staticmethod
    def handle_list_block(
        is_ulist,
        parser_state,
        did_process,
        was_container_start,
        position_marker,
        extracted_whitespace,
        adj_ws,
        stack_bq_count,
        this_bq_count,
        removed_chars_at_start,
        current_container_blocks,
    ):
        """
        Handle the processing of a ulist block.
        """
        (
            requeue_line_info,
            end_of_ulist_start_index,
            container_level_tokens,
            adjusted_text_to_parse,
        ) = (None, -1, [], position_marker.text_to_parse)

        if not did_process:

            if is_ulist:
                is_start_fn = ListBlockProcessor.is_ulist_start
                create_token_fn = ListBlockProcessor.__handle_list_block_unordered
            else:
                is_start_fn = ListBlockProcessor.is_olist_start
                create_token_fn = ListBlockProcessor.__handle_list_block_ordered

            (
                started_ulist,
                end_of_ulist_start_index,
                index,
                number_of_digits,
            ) = is_start_fn(
                parser_state,
                position_marker.text_to_parse,
                position_marker.index_number,
                extracted_whitespace,
                False,
                adj_ws=adj_ws,
            )
            if started_ulist:
                POGGER.debug("clt>>ulist-start")
                removed_chars_at_start = 0

                (
                    indent_level,
                    remaining_whitespace,
                    ws_after_marker,
                    after_marker_ws_index,
                    ws_before_marker,
                    container_level_tokens,
                    stack_bq_count,
                ) = ListBlockProcessor.__pre_list(
                    parser_state,
                    position_marker.text_to_parse,
                    index,
                    extracted_whitespace,
                    number_of_digits,
                    stack_bq_count,
                    this_bq_count,
                    adj_ws=adj_ws,
                )

                POGGER.debug(
                    "total=$;ws-before=$;ws_after=$;start_index=$",
                    indent_level,
                    ws_before_marker,
                    ws_after_marker,
                    position_marker.index_number,
                )
                if indent_level >= 0:
                    new_token, new_stack = create_token_fn(
                        position_marker,
                        indent_level,
                        extracted_whitespace,
                        ws_before_marker,
                        ws_after_marker,
                        index,
                    )

                    (
                        new_container_level_tokens,
                        adjusted_text_to_parse,
                        requeue_line_info,
                    ) = ListBlockProcessor.__post_list(
                        parser_state,
                        new_stack,
                        new_token,
                        position_marker.text_to_parse,
                        remaining_whitespace,
                        after_marker_ws_index,
                        indent_level,
                        current_container_blocks,
                        position_marker,
                    )
                    if new_container_level_tokens:
                        container_level_tokens.extend(new_container_level_tokens)
                    did_process, was_container_start = True, True
        return (
            did_process,
            was_container_start,
            end_of_ulist_start_index,
            adjusted_text_to_parse,
            container_level_tokens,
            removed_chars_at_start,
            requeue_line_info,
        )
        # pylint: enable=too-many-locals, too-many-arguments

    @staticmethod
    # pylint: disable=too-many-arguments
    def __handle_list_block_unordered(
        position_marker,
        indent_level,
        extracted_whitespace,
        ws_before_marker,
        ws_after_marker,
        index,
    ):
        _ = index

        new_token = UnorderedListStartMarkdownToken(
            position_marker.text_to_parse[position_marker.index_number],
            indent_level,
            extracted_whitespace,
            position_marker,
        )
        return new_token, UnorderedListStackToken(
            indent_level,
            position_marker.text_to_parse[position_marker.index_number],
            ws_before_marker,
            ws_after_marker,
            position_marker.index_number,
            new_token,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    # pylint: disable=too-many-arguments
    def __handle_list_block_ordered(
        position_marker,
        indent_level,
        extracted_whitespace,
        ws_before_marker,
        ws_after_marker,
        index,
    ):
        new_token = OrderedListStartMarkdownToken(
            position_marker.text_to_parse[index],
            position_marker.text_to_parse[position_marker.index_number : index],
            indent_level,
            extracted_whitespace,
            position_marker,
        )
        return new_token, OrderedListStackToken(
            indent_level,
            position_marker.text_to_parse[position_marker.index_number : index + 1],
            ws_before_marker,
            ws_after_marker,
            position_marker.index_number,
            new_token,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-statements, too-many-locals
    @staticmethod
    def list_in_process(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        ind,
    ):
        """
        Handle the processing of a line where there is a list in process.
        """
        (
            container_level_tokens,
            requested_list_indent,
            before_ws_length,
            leading_space_length,
        ) = (
            [],
            parser_state.token_stack[ind].indent_level,
            parser_state.token_stack[ind].ws_before_marker,
            ParserHelper.calculate_length(extracted_whitespace),
        )

        POGGER.debug("!!!!!FOUND>>$", parser_state.token_stack[ind])
        POGGER.debug("!!!!!FOUND>>$", parser_state.token_stack[ind].extra_data)
        POGGER.debug("!!!!!ALL>>$", parser_state.token_stack)
        POGGER.debug("!!!!!ALL>>$", parser_state.token_document)

        POGGER.debug(
            "!!!!!requested_list_indent>>$,before_ws=$",
            requested_list_indent,
            before_ws_length,
        )

        started_ulist, _, _, _ = ListBlockProcessor.is_ulist_start(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            True,
        )
        started_olist, _, _, _ = ListBlockProcessor.is_olist_start(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            True,
        )

        allow_list_continue = (
            (not parser_state.token_document[-1].is_blank_line)
            if leading_space_length >= 4 and (started_ulist or started_olist)
            else True
        )

        POGGER.debug(
            "leading_space_length>>$>>requested_list_indent>>$>>is_in_paragraph>>$",
            leading_space_length,
            requested_list_indent,
            parser_state.token_stack[-1].is_paragraph,
        )

        used_indent = None

        if leading_space_length >= requested_list_indent and allow_list_continue:

            POGGER.debug("before>>$>>", line_to_parse)
            (
                line_to_parse,
                used_indent,
            ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                line_to_parse,
                start_index,
                extracted_whitespace,
                leading_space_length,
                requested_list_indent,
            )
            POGGER.debug(
                "after>>$>>$>>",
                line_to_parse,
                used_indent,
            )
        else:
            POGGER.debug(
                "requested_list_indent>>$<<",
                requested_list_indent,
            )
            original_requested_list_indent = requested_list_indent
            requested_list_indent = requested_list_indent - before_ws_length
            POGGER.debug(
                "leading_space_length>>$>>adj requested_list_indent>>$>>$<<",
                leading_space_length,
                requested_list_indent,
                parser_state.token_stack[-1].is_paragraph,
            )

            # This needs to be in place to prevent a thematic break after a paragraph
            # within a list from being misinterpretted as a SetExt Heading.
            is_theme_break, _ = LeafBlockProcessor.is_thematic_break(
                line_to_parse,
                start_index,
                extracted_whitespace,
                skip_whitespace_check=True,
            )
            POGGER.debug("is_theme_break>>$", is_theme_break)

            if (
                parser_state.token_stack[-1].is_paragraph
                and leading_space_length >= requested_list_indent
                and allow_list_continue
                and not is_theme_break
            ):
                POGGER.debug(
                    "1>>line_to_parse>>$>>",
                    line_to_parse,
                )
                (
                    line_to_parse,
                    used_indent,
                ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                    line_to_parse,
                    start_index,
                    extracted_whitespace,
                    leading_space_length,
                    original_requested_list_indent,
                )
                POGGER.debug(
                    ">>line_to_parse>>$>>",
                    line_to_parse,
                )
                POGGER.debug(">>used_indent>>$>>", used_indent)
            else:
                POGGER.debug(
                    "2>>line_to_parse>>$>>",
                    line_to_parse,
                )
                container_level_tokens = ListBlockProcessor.__check_for_list_closures(
                    parser_state,
                    line_to_parse,
                    start_index,
                    extracted_whitespace,
                    ind,
                    leading_space_length,
                )

                POGGER.debug(
                    "2>>__check_for_list_closures>>$>>",
                    container_level_tokens,
                )
                POGGER.debug(
                    "2>>parser_state.token_stack>>$>>",
                    parser_state.token_stack,
                )
                POGGER.debug(
                    "2>>ind>>$>>",
                    ind,
                )

                found_owning_list = None
                if container_level_tokens:
                    (
                        did_find,
                        last_list_index,
                    ) = LeafBlockProcessor.check_for_list_in_process(parser_state)
                    POGGER.debug(
                        "2>>did_find>>$>>$>>",
                        did_find,
                        last_list_index,
                    )
                    if did_find:
                        ind = last_list_index
                        found_owning_list = parser_state.token_stack[ind]
                else:
                    assert parser_state.token_stack[ind].is_list
                    found_owning_list = parser_state.token_stack[ind]

                if found_owning_list:
                    POGGER.debug(">>in list>>")
                    requested_list_indent = (
                        (found_owning_list.last_new_list_token.indent_level)
                        if found_owning_list.last_new_list_token
                        else found_owning_list.indent_level
                    )
                    POGGER.debug(">>line_to_parse>>$>>", line_to_parse)
                    POGGER.debug(">>extracted_whitespace>>$<<", extracted_whitespace)
                    POGGER.debug(">>start_index>>$", start_index)
                    POGGER.debug(">>requested_list_indent>>$", requested_list_indent)
                    POGGER.debug(">>before_ws_length>>$", before_ws_length)
                    (
                        line_to_parse,
                        used_indent,
                    ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                        line_to_parse,
                        start_index,
                        extracted_whitespace,
                        leading_space_length,
                        requested_list_indent,
                    )
                    POGGER.debug(">>line_to_parse>>$", line_to_parse)
                    POGGER.debug(">>used_indent>>$<<", used_indent)

        POGGER.debug(">>used_indent>>$<<", used_indent)
        if used_indent is not None:
            POGGER.debug(
                ">>adj_before>>$<<",
                parser_state.token_stack[ind].matching_markdown_token,
            )
            parser_state.token_stack[ind].matching_markdown_token.add_leading_spaces(
                used_indent
            )
            POGGER.debug(
                ">>adj_after>>$<<",
                parser_state.token_stack[ind].matching_markdown_token,
            )
        return container_level_tokens, line_to_parse, used_indent

    # pylint: enable=too-many-statements, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __pre_list(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        marker_width_minus_one,
        stack_bq_count,
        this_bq_count,
        adj_ws,
    ):
        """
        Handle the processing of the first part of the list.
        """

        (
            after_marker_ws_index,
            after_marker_whitespace,
        ) = ParserHelper.extract_whitespace(line_to_parse, start_index + 1)
        ws_after_marker, ws_before_marker, line_to_parse_size = (
            ParserHelper.calculate_length(
                after_marker_whitespace, start_index=start_index + 1
            ),
            ParserHelper.calculate_length(extracted_whitespace),
            len(line_to_parse),
        )
        POGGER.debug(
            "after-marker>>$>>total=$", after_marker_whitespace, ws_after_marker
        )
        POGGER.debug(
            "--ws_before_marker>>$>>marker_width_minus_one>>$",
            ws_before_marker,
            marker_width_minus_one,
        )
        POGGER.debug("--$--$", start_index, start_index + 1)
        POGGER.debug(">>>>>XX>>$>>$<<", after_marker_ws_index, line_to_parse_size)
        if after_marker_ws_index == line_to_parse_size and ws_after_marker:
            indent_level, remaining_whitespace, ws_after_marker = (
                2 + marker_width_minus_one + len(adj_ws),
                ws_after_marker,
                0,
            )
        else:
            if after_marker_ws_index == line_to_parse_size and ws_after_marker == 0:
                ws_after_marker += 1

            indent_level = (
                ws_before_marker + 1 + ws_after_marker + marker_width_minus_one
            )
            if ws_after_marker > 4:
                indent_level, remaining_whitespace, ws_after_marker = (
                    indent_level - ws_after_marker + 1,
                    ws_after_marker - 1,
                    1,
                )
            else:
                remaining_whitespace = 0

        if (
            parser_state.token_stack[-1].is_html_block
            or parser_state.token_stack[-1].is_fenced_code_block
        ):
            did_find, _ = LeafBlockProcessor.check_for_list_in_process(parser_state)
            if not did_find:
                indent_level = -1
                after_marker_ws_index = -1
                POGGER.debug("BAIL!")

        (
            container_level_tokens,
            stack_bq_count,
        ) = ListBlockProcessor.__handle_list_nesting(
            parser_state, stack_bq_count, this_bq_count
        )
        POGGER.debug(
            "ws_after_marker>>$<<indent_level<<$<<rem<<$<<",
            ws_after_marker,
            indent_level,
            remaining_whitespace,
        )
        return (
            indent_level,
            remaining_whitespace,
            ws_after_marker,
            after_marker_ws_index,
            ws_before_marker,
            container_level_tokens,
            stack_bq_count,
        )
        # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __handle_list_nesting(parser_state, stack_bq_count, this_bq_count):
        """
        Resolve any nesting issues with block quotes.
        """
        POGGER.debug(
            ">>stack_bq_count>>$>>this_bq_count>>$",
            stack_bq_count,
            this_bq_count,
        )
        container_level_tokens = []
        while this_bq_count < stack_bq_count:

            assert not container_level_tokens
            last_block_index = parser_state.find_last_block_quote_on_stack()
            container_level_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_block_index,
                include_block_quotes=True,
                include_lists=True,
            )
            POGGER.debug("container_level_tokens>>$", container_level_tokens)
            stack_bq_count -= 1
        return container_level_tokens, stack_bq_count

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __post_list(
        parser_state,
        new_stack,
        new_token,
        line_to_parse,
        remaining_whitespace,
        after_marker_ws_index,
        indent_level,
        current_container_blocks,
        position_marker,
    ):
        """
        Handle the processing of the last part of the list.
        """

        POGGER.debug("new_stack>>$", new_stack)
        POGGER.debug("indent_level>>$", indent_level)

        emit_item, emit_li = True, True
        did_find, last_list_index = LeafBlockProcessor.check_for_list_in_process(
            parser_state
        )
        if did_find:
            (
                container_level_tokens,
                emit_li,
                requeue_line_info,
            ) = ListBlockProcessor.__close_required_lists_after_start(
                parser_state,
                last_list_index,
                new_stack,
                current_container_blocks,
            )
            if requeue_line_info:
                return None, None, requeue_line_info
            emit_item = False
        else:
            POGGER.debug(
                "NOT list-in-process>>$",
                parser_state.token_stack[last_list_index],
            )
            container_level_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state, was_forced=True
            )
        POGGER.debug("container_level_tokens>>$", container_level_tokens)

        POGGER.debug("__post_list>>before>>$", container_level_tokens)
        if emit_item or not emit_li:
            POGGER.debug("__post_list>>adding>>$", new_token)
            parser_state.token_stack.append(new_stack)
            container_level_tokens.append(new_token)
        else:
            POGGER.debug("__post_list>>new list item>>")
            assert emit_li
            ListBlockProcessor.__post_list_use_new_list_item(
                parser_state,
                new_token,
                container_level_tokens,
                indent_level,
                position_marker,
            )
        POGGER.debug(
            "__post_list>>rem>>$>>after_in>>$",
            remaining_whitespace,
            after_marker_ws_index,
        )
        POGGER.debug("__post_list>>after>>$", container_level_tokens)

        parser_state.set_no_para_start_if_empty()
        return (
            container_level_tokens,
            f"{ParserHelper.repeat_string(ParserHelper.space_character, remaining_whitespace)}{line_to_parse[after_marker_ws_index:]}",
            None,
        )
        # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __post_list_use_new_list_item(
        parser_state, new_token, container_level_tokens, indent_level, position_marker
    ):
        POGGER.debug("instead of-->$", new_token)

        top_stack_item = parser_state.token_stack[-1]
        assert top_stack_item.is_list

        list_start_content = ""
        if top_stack_item.is_ordered_list:
            list_start_content = new_token.list_start_content
            POGGER.debug("ordered->start-->$", new_token.list_start_content)
        else:
            POGGER.debug("unordered->start-->")

        new_token = NewListItemMarkdownToken(
            indent_level,
            position_marker,
            new_token.extracted_whitespace,
            list_start_content,
        )
        top_stack_item.set_last_new_list_token(new_token)
        container_level_tokens.append(new_token)

    @staticmethod
    def __close_required_lists_after_start(
        parser_state,
        last_list_index,
        new_stack,
        current_container_blocks,
    ):
        """
        After a list start, check to see if any others need closing.
        """
        POGGER.debug("list-in-process>>$", parser_state.token_stack[last_list_index])
        (
            container_level_tokens,
            requeue_line_info,
        ) = parser_state.close_open_blocks_fn(
            parser_state,
            until_this_index=last_list_index + 1,
            caller_can_handle_requeue=True,
        )
        if requeue_line_info and requeue_line_info.lines_to_requeue:
            POGGER.debug(
                "__close_required_lists_after_start>>lines_to_requeue>>$",
                requeue_line_info.lines_to_requeue,
            )
            POGGER.debug(
                "__close_required_lists_after_start>>parser_state.original_line_to_parse>>$",
                parser_state.original_line_to_parse,
            )
            POGGER.debug(
                "__close_required_lists_after_start>>token_stack>>$",
                parser_state.token_stack,
            )
            POGGER.debug(
                "__close_required_lists_after_start>>token_document>>$",
                parser_state.token_document,
            )
            assert not requeue_line_info.lines_to_requeue[0]
            requeue_line_info.lines_to_requeue[0] = parser_state.original_line_to_parse
            POGGER.debug(
                "__close_required_lists_after_start>>lines_to_requeue>>$",
                requeue_line_info.lines_to_requeue,
            )
            return None, None, requeue_line_info

        POGGER.debug("old-stack>>$<<", container_level_tokens)
        repeat_check, emit_li = True, False
        while repeat_check:
            POGGER.debug("start")
            repeat_check = False
            (
                do_not_emit,
                emit_li,
                extra_tokens,
                last_list_index,
            ) = ListBlockProcessor.__are_list_starts_equal(
                parser_state,
                last_list_index,
                new_stack,
                current_container_blocks,
            )
            POGGER.debug("extra_tokens>>$", extra_tokens)
            container_level_tokens.extend(extra_tokens)
            if do_not_emit:
                POGGER.debug("post_list>>don't emit")
                (
                    did_find,
                    last_list_index,
                ) = LeafBlockProcessor.check_for_list_in_process(parser_state)
                POGGER.debug(
                    "did_find>>$--last_list_index--$",
                    did_find,
                    last_list_index,
                )
                assert did_find
                POGGER.debug(
                    "ARE-EQUAL>>stack>>$>>new>>$",
                    parser_state.token_stack[last_list_index],
                    new_stack,
                )
                if not (
                    parser_state.token_stack[last_list_index].type_name
                    == new_stack.type_name
                    or new_stack.start_index
                    > parser_state.token_stack[last_list_index].start_index
                ):
                    repeat_check = True
            else:
                POGGER.debug("post_list>>close open blocks and emit")
                close_tokens, _ = parser_state.close_open_blocks_fn(
                    parser_state, until_this_index=last_list_index, include_lists=True
                )
                assert close_tokens
                container_level_tokens.extend(close_tokens)

                (
                    did_find,
                    last_list_index,
                ) = LeafBlockProcessor.check_for_list_in_process(parser_state)
                POGGER.debug(
                    "did_find>>$--last_list_index--$",
                    did_find,
                    last_list_index,
                )
                if did_find:
                    POGGER.debug(
                        "ARE-EQUAL>>stack>>$>>new>>$",
                        parser_state.token_stack[last_list_index],
                        new_stack,
                    )
                    if (
                        new_stack.indent_level
                        <= parser_state.token_stack[last_list_index].indent_level
                    ):
                        repeat_check = True
        return container_level_tokens, emit_li, None

    @staticmethod
    def __are_list_starts_equal(
        parser_state,
        last_list_index,
        new_stack,
        current_container_blocks,
    ):
        """
        Check to see if the list starts are equal, and hence a continuation of
        the current list.
        """

        balancing_tokens = []

        POGGER.debug(
            "ARE-EQUAL>>stack>>$>>new>>$",
            parser_state.token_stack[last_list_index],
            new_stack,
        )
        if parser_state.token_stack[last_list_index] == new_stack:
            balancing_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_list_index,
                include_block_quotes=True,
            )
            return True, True, balancing_tokens, last_list_index

        document_token_index = len(parser_state.token_document) - 1
        while document_token_index >= 0 and not (
            parser_state.token_document[document_token_index].is_any_list_token
        ):
            document_token_index -= 1
        assert document_token_index >= 0

        POGGER.debug(
            "ARE-EQUAL>>Last_List_token=$",
            parser_state.token_document[document_token_index],
        )
        old_start_index, old_last_marker_character, current_start_index = (
            parser_state.token_document[document_token_index].indent_level,
            parser_state.token_stack[last_list_index].list_character[-1],
            new_stack.ws_before_marker,
        )
        POGGER.debug(
            "old>>$>>$",
            parser_state.token_stack[last_list_index].extra_data,
            old_last_marker_character,
        )
        POGGER.debug("new>>$>>$", new_stack.extra_data, new_stack.list_character[-1])
        if (
            parser_state.token_stack[last_list_index].type_name == new_stack.type_name
            and old_last_marker_character == new_stack.list_character[-1]
        ):
            do_not_emit, emit_li = ListBlockProcessor.__process_eligible_list_start(
                parser_state,
                balancing_tokens,
                current_start_index,
                old_start_index,
                current_container_blocks,
            )
            return do_not_emit, emit_li, balancing_tokens, last_list_index

        POGGER.debug("SUBLIST WITH DIFFERENT")
        POGGER.debug("are_list_starts_equal>>ELIGIBLE!!!")
        POGGER.debug(
            "are_list_starts_equal>>current_start_index>>$>>old_start_index>>$",
            current_start_index,
            old_start_index,
        )
        if current_start_index >= old_start_index:
            POGGER.debug("are_list_starts_equal>>True")
            return True, False, balancing_tokens, last_list_index

        POGGER.debug("are_list_starts_equal>>False")
        POGGER.debug(">>$", parser_state.token_stack)
        return False, False, balancing_tokens, last_list_index

    @staticmethod
    def __process_eligible_list_start(
        parser_state,
        balancing_tokens,
        current_start_index,
        old_start_index,
        current_container_blocks,
    ):
        POGGER.debug("are_list_starts_equal>>ELIGIBLE!!!")
        POGGER.debug(
            "are_list_starts_equal>>current_start_index>>$>>old_start_index>>$",
            current_start_index,
            old_start_index,
        )
        if current_start_index < old_start_index:

            POGGER.debug("current_container_blocks>>$", current_container_blocks)
            if len(current_container_blocks) > 1:
                POGGER.debug("current_container_blocks-->$", parser_state.token_stack)
                last_stack_depth = parser_state.token_stack[-1].ws_before_marker
                while current_start_index < last_stack_depth:
                    last_stack_index = parser_state.token_stack.index(
                        parser_state.token_stack[-1]
                    )
                    close_tokens, _ = parser_state.close_open_blocks_fn(
                        parser_state,
                        until_this_index=last_stack_index,
                        include_lists=True,
                    )
                    assert close_tokens
                    balancing_tokens.extend(close_tokens)
                    POGGER.debug("close_tokens>>$", close_tokens)
                    last_stack_depth = parser_state.token_stack[-1].ws_before_marker

            return True, True

        POGGER.debug("are_list_starts_equal>>True")
        return True, False

    @staticmethod
    def __adjust_line_for_list_in_process(
        line_to_parse,
        start_index,
        leading_space,
        leading_space_length,
        requested_list_indent,
    ):
        """
        Alter the current line to better represent the current level of lists.
        """
        remaining_indent = leading_space_length - requested_list_indent
        POGGER.debug(
            "enough ws to continue; line($),start_index($),leading_space($)",
            line_to_parse,
            start_index,
            leading_space,
        )
        POGGER.debug(
            "enough ws to continue; lsl($)-rsi($)=ri($)",
            leading_space_length,
            requested_list_indent,
            remaining_indent,
        )
        removed_whitespace = (
            ParserHelper.tab_character
            if ParserHelper.tab_character in leading_space
            else leading_space[0:requested_list_indent]
        )
        return (
            f"{ParserHelper.repeat_string(ParserHelper.space_character, remaining_indent)}{line_to_parse[start_index:]}",
            removed_whitespace,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __check_for_list_closures(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        ind,
        leading_space_length,
    ):
        """
        Check to see if the list in progress and the level of lists shown require
        the closing of some of the sublists.
        """
        container_level_tokens = []
        POGGER.debug("ws(naa)>>line_to_parse>>$<<", line_to_parse)
        POGGER.debug("ws(naa)>>stack>>$", parser_state.token_stack)
        POGGER.debug("ws(naa)>>tokens>>$", parser_state.token_document)

        is_leaf_block_start = LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
            parser_state, line_to_parse, start_index, extracted_whitespace
        )
        if (
            not parser_state.token_stack[-1].is_paragraph or is_leaf_block_start
        ) and not parser_state.token_stack[-1].was_link_definition_started:
            POGGER.debug("ws (normal and adjusted) not enough to continue")

            POGGER.debug("lsl $", leading_space_length)
            POGGER.debug("lsl $", parser_state.token_stack[ind])
            search_index = ind
            POGGER.debug(
                "lsl $>$",
                search_index,
                parser_state.token_stack[search_index - 1],
            )
            while (
                parser_state.token_stack[search_index - 1].is_list
                and parser_state.token_stack[search_index - 1].indent_level
                > leading_space_length
            ):
                search_index -= 1
                POGGER.debug(
                    "lsl $>$",
                    search_index,
                    parser_state.token_stack[search_index],
                )

            POGGER.debug("lsl $", parser_state.token_stack[search_index])

            container_level_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state, until_this_index=search_index, include_lists=True
            )
            POGGER.debug("container_level_tokens>$>", container_level_tokens)
        return container_level_tokens

    # pylint: enable=too-many-arguments



"""
Tests for the various is_character* functions.
"""
from pymarkdown.parser_helper import ParserHelper


def test_is_character_at_index_whitespace_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_whitespace_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_whitespace_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_whitespace_with_whitespace():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = " "
    start_index = 0
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_whitespace_with_whitespace_at_end():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test "
    start_index = len(input_string) - 1
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_whitespace_without_whitespace():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_with_whitespace():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = " "
    start_index = 0
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_with_whitespace_at_end():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test "
    start_index = len(input_string) - 1
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_whitespace_without_whitespace():
    """
    Make sure that a string with whitespace at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_not_whitespace(
        input_string, start_index
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_whitespace():
    """
    Make sure that a string with one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    valid_characters = "abc"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_whitespace2():
    """
    Make sure that a string with another one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "c"
    start_index = 0
    valid_characters = "abc"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_with_character_at_end():
    """
    Make sure that a string with one of the characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test!"
    start_index = len(input_string) - 1
    valid_characters = "abc!"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_one_of_without_whitespace():
    """
    Make sure that a string without any characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 0
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_whitespace():
    """
    Make sure that a string with one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_whitespace2():
    """
    Make sure that a string with another one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "c"
    start_index = 0
    valid_characters = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_with_character_at_end():
    """
    Make sure that a string with one of the characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test!"
    start_index = len(input_string) - 1
    valid_characters = "abc!"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_one_of_without_whitespace():
    """
    Make sure that a string without any characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 0
    valid_characters = "abc"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_not_one_of(
        input_string, start_index, valid_characters
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_with_whitespace():
    """
    Make sure that a string with one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    valid_character = "a"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_with_character_at_end():
    """
    Make sure that a string with one of the characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test!"
    start_index = len(input_string) - 1
    valid_character = "!"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_without_whitespace():
    """
    Make sure that a string without any characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 0
    valid_character = "b"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = len(input_string)
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_with_whitespace():
    """
    Make sure that a string with one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "a"
    start_index = 0
    valid_character = "a"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_with_character_at_end():
    """
    Make sure that a string with one of the characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test!"
    start_index = len(input_string) - 1
    valid_character = "!"
    expected_output = False

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_is_character_at_index_not_without_whitespace():
    """
    Make sure that a string without any characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 0
    valid_character = "b"
    expected_output = True

    # Act
    actual_output = ParserHelper.is_character_at_index_not(
        input_string, start_index, valid_character
    )

    # Assert
    assert expected_output == actual_output


def test_are_characters_at_index_with_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    start_index = 0
    sequence_to_look_for = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output


def test_are_characters_at_index_with_low_index():
    """
    Make sure that a string with a low index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = -1
    sequence_to_look_for = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output


def test_are_characters_at_index_with_high_index():
    """
    Make sure that a string with a high index is handled properly.
    """

    # Arrange
    input_string = "this is a abc"
    sequence_to_look_for = "abc"
    start_index = len(input_string) - len(sequence_to_look_for) + 1
    expected_output = False

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output


def test_are_characters_at_index_with_present():
    """
    Make sure that a string with one of the characters present at the index is handled properly.
    """

    # Arrange
    input_string = "abc"
    start_index = 0
    sequence_to_look_for = "abc"
    expected_output = True

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output


def test_are_characters_at_index_with_character_at_end():
    """
    Make sure that a string with one of the characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a abc"
    sequence_to_look_for = "abc"
    start_index = len(input_string) - len(sequence_to_look_for)
    expected_output = True

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output


def are_characters_at_index_with_not_present():
    """
    Make sure that a string without any characters at the index is handled properly.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 6
    sequence_to_look_for = "abc"
    expected_output = False

    # Act
    actual_output = ParserHelper.are_characters_at_index(
        input_string, start_index, sequence_to_look_for
    )

    # Assert
    assert expected_output == actual_output



"""
Module to implement a plugin that looks for hard tabs in the files.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd006(Plugin):
    """
    Class to implement a plugin that looks for hard tabs in the files.
    """

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            # bullet, ul, indentation
            plugin_name="ul-start-left",
            plugin_id="MD006",
            plugin_enabled_by_default=False,
            plugin_description="Consider starting bulleted lists at the beginning of the line",
            plugin_version="0.0.0",
            plugin_interface_version=1,
        )  # https://github.com/DavidAnson/markdownlint/blob/master/doc/Rules.md#md006---consider-starting-bulleted-lists-at-the-beginning-of-the-line



"""
Setup file for the PyMarkdown project.
"""

import runpy
from distutils import util

from setuptools import setup


def parse_requirements():
    lineiter = (line.strip() for line in open("install-requirements.txt", "r"))
    return [line for line in lineiter if line and not line.startswith("#")]


def get_semantic_version():
    version_meta = runpy.run_path("./pymarkdown/version.py")
    return version_meta["__version__"]


def load_readme_file():
    with open("README.md", "r") as readme_file:
        return readme_file.read()


def ensure_scripts(linux_scripts):
    """
    Creates the proper script names required for each platform (taken from PyLint)
    """
    if util.get_platform()[:3] == "win":
        return linux_scripts + [script + ".bat" for script in linux_scripts]
    return linux_scripts


AUTHOR = "Jack De Winter"
AUTHOR_EMAIL = "jack.de.winter@outlook.com"
PROJECT_URL = "https://github.com/jackdewinter/pymarkdown"

PACKAGE_NAME = "pymarkdownlnt"
SEMANTIC_VERSION = get_semantic_version()
MINIMUM_PYTHON_VERSION = "3.8.0"

ONE_LINE_DESCRIPTION = "A GitHub Flavored Markdown compliant Markdown linter."
LONG_DESCRIPTION = load_readme_file()
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"

KEYWORDS = ["markdown", "linter", "markdown linter"]
PROJECT_CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Programming Language :: Python :: 3.8",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]

PACKAGE_MODULES = [
    "pymarkdown",
    "pymarkdown.extensions",
    "pymarkdown.plugins",
    "pymarkdown.resources",
]
PACKAGE_SCRIPTS = ensure_scripts(["scripts/pymarkdown"])

setup(
    name=PACKAGE_NAME,
    version=SEMANTIC_VERSION,
    python_requires=">=" + MINIMUM_PYTHON_VERSION,
    install_requires=parse_requirements(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=AUTHOR_EMAIL,
    url=PROJECT_URL,
    description=ONE_LINE_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    keywords=KEYWORDS,
    classifiers=PROJECT_CLASSIFIERS,
    scripts=PACKAGE_SCRIPTS,
    packages=PACKAGE_MODULES,
    data_files=[
        (
            "Lib/site-packages/pymarkdown/resources",
            ["pymarkdown/resources/entities.json"],
        )
    ],
)
